"""Bumo configuration variables."""

from build123d import Color
from .colors import ColorPalette


COLOR_PALETTE = ColorPalette.VIRIDIS
"The color palette to use when auto_color is enabled."

DEBUG_ALPHA = 0.2
"The alpha value used for translucent shapes in debug mode."

DEFAULT_COLOR = Color("orange")
"The default color to be used when a color is passed to a mutation."

DEFAULT_DEBUG_COLOR = Color("red")
"The default color to be used when using the debug mode."

INFO_COLOR = True
"Set to False to disable terminal colors in the info table."

INFO_TABLE_FORMAT = "fancy_outline"
"""The table format used in the info table. See
https://github.com/astanin/python-tabulate?tab=readme-ov-file#table-format
"""

COLUMNS_MUTATIONS = ["idx", "label", "type", "f+", "f~", "f-", "e+", "e~", "e-"]
""""The columns to display in mutations info tables, among:
idx, label, type, color_hex, color_name, f+, f~, f-, e+, e~, e-."""

COLUMNS_SHAPES = ["hash", "type", "area", "position", "orientation"]
""""The columns to display in shapes info tables, among:
hash, type, area, color_hex, color_name."""



from enum import Enum

import build123d as _

from . import config
from .colors import ColorLike, cast_color


class ModeType(Enum):
    FIXED = 1
    AUTO = 2
    DEBUG = 3
    DEFAULT = 4


class Mode:
    def __init__(self, mode_type: ModeType, color: _.Color) -> None:
        self.mode_type = mode_type
        self.color = color

    def get_color(self, auto_color: _.Color):
        if self.mode_type == ModeType.AUTO:
            return auto_color

        if self.mode_type == ModeType.DEFAULT:
            return config.DEFAULT_COLOR

        return self.color


def cast_mode(mode: Mode | ColorLike) -> Mode:
    if isinstance(mode, Mode):
        return mode
    return Mode(ModeType.FIXED, cast_color(mode))


AUTO = Mode(ModeType.AUTO, config.DEFAULT_COLOR)
DEBUG = Mode(ModeType.DEBUG, config.DEFAULT_DEBUG_COLOR)



"""Module containing the Builder class."""
from __future__ import annotations
from os import PathLike
from sys import stdout
from typing import Iterable

import build123d as _
from tabulate import tabulate

from .mutation import Mutation
from .mode import Mode, ModeType, cast_mode, AUTO, DEBUG
from .colors import color_to_str, get_rvb, ColorLike
from .shapes import Hash, ShapeList, add_shape_hash
from . import config


class Builder:
    """A class used to manipulate Build123d objects that keeps track of each
    mutation and manage shape colors."""

    def __init__(self):

        self.object = _.Part(None)
        self.mutations: list[Mutation] = []
        self.faces_modes: dict[Hash, Mode] = {}
        self.faces_dict: dict[Hash, _.Face] = {}

    def __getitem__(self, mut_idx: int):
        return self.mutations[mut_idx]

    def get_faces_colors(self) -> dict[Hash, _.Color]:
        """Build a dict containing for each face hash, its actual color."""

        palette = config.COLOR_PALETTE.build_palette(len(self.mutations))
        faces_mutations = self.get_faces_mutations()
        faces_last = self.last.faces.hashes()

        def get_color(face_hash):
            mode = self.faces_modes[face_hash]
            mut_idx = faces_mutations[face_hash]
            return mode.get_color(palette[mut_idx])

        faces_debug = {
            face_hash: mode.color
            for face_hash, mode in self.faces_modes.items()
            if mode.mode_type == ModeType.DEBUG
        }

        if faces_debug:
            ghost_colors = {
                fg_h: _.Color(*get_rvb(get_color(fg_h)), config.DEBUG_ALPHA)
                for fg_h in faces_last
            }
            return {
                **ghost_colors,
                **faces_debug,
            }

        return {fl_h: get_color(fl_h) for fl_h in faces_last}

    def __call__(self) -> list[_.Face]:
        if not self.mutations:
            raise ValueError("No mutation to show.")

        faces_colors = self.get_faces_colors()
        faces_to_show: list[_.Face] = []

        for face_hash, face_color in faces_colors.items():
            face = self.faces_dict[face_hash]
            face.color = face_color
            face.label = face_hash[:6]
            faces_to_show.append(face)

        return faces_to_show

    def __iadd__(self, part: Builder | _.Part | tuple[Builder | _.Part, Mode | ColorLike]):
        if isinstance(part, tuple):
            self.add(*part)
        else:
            self.add(part)
        return self

    def __isub__(self, part: Builder | _.Part | tuple[Builder | _.Part, Mode | ColorLike]):
        if isinstance(part, tuple):
            self.sub(*part)
        else:
            self.sub(part)
        return self

    def __imul__(self, location: _.Location | tuple[_.Location, Mode | ColorLike]):
        if isinstance(location, tuple):
            self.move(*location)
        else:
            self.move(location)
        return self

    def __iand__(self, part: Builder | _.Part | tuple[Builder | _.Part, Mode | ColorLike]):
        if isinstance(part, tuple):
            self.intersect(*part)
        else:
            self.intersect(part)
        return self

    @property
    def last(self):
        """Return the last mutation."""
        return self.mutations[-1]

    def get_faces_mutations(self) -> dict[Hash, int]:
        """Return a dictionnary containing for each face hash, the mutation that
        created the face."""

        # TODO: store dict in builder and apply on each mutation?
        faces_mutations: dict[Hash, int] = {}

        for mutation in self.mutations:

            for face_ad in mutation.faces_added:
                faces_mutations[face_ad.label] = (
                    faces_mutations[mutation.faces_alias[face_ad.label]]
                    if mutation.faces_alias
                    else mutation.index
                )

            rm_muts = {faces_mutations[face_rm.label] for face_rm in mutation.faces_removed}

            if len(rm_muts) == 1:
                rm_color = rm_muts.pop()

                for face_al in mutation.faces_altered:
                    faces_mutations[face_al.label] = rm_color
            else:
                for face_al in mutation.faces_altered:
                    for face_rm in mutation.faces_removed:
                        if Mutation.is_altered_faces(face_al, face_rm):
                            rm_mut_idx = faces_mutations[face_rm.label]
                            faces_mutations[face_al.label] = rm_mut_idx

        return faces_mutations

    def get_mutation(self, mutation_id: str) -> Mutation:
        """Return the mutation identified by the given id."""

        for mutation in self.mutations:
            if mutation.id == mutation_id:
                return mutation
        raise KeyError

    @classmethod
    def _cast_faces(cls, faces: Iterable[_.Face] | _.Face) -> ShapeList[_.Face]:
        """Cast the given faces to a FaceDict."""

        if isinstance(faces, ShapeList):
            return faces

        if isinstance(faces, _.Face):
            return ShapeList([faces])

        return ShapeList(faces)

    @classmethod
    def _cast_edges(cls, edges: Iterable[_.Edge] | _.Edge) -> ShapeList[_.Edge]:
        """Cast the given edges an EdgeDict."""

        if isinstance(edges, ShapeList):
            return edges

        if isinstance(edges, _.Edge):
            return ShapeList([edges])

        return ShapeList(edges)

    @classmethod
    def _cast_part(cls, part: Builder | _.Part) -> _.Part:
        """Cast an EdgeListLike to a Edge iterable."""
        return part if isinstance(part, _.Part) else part.object

    def mutate(
            self,
            name: str,
            obj: _.Part,
            mode: Mode | ColorLike,
            faces_alias: dict[Hash, Hash] | None=None
        ) -> Mutation:
        """Base mutation: mutate the current object to the given one by applying
        a mutation with the given name, color and debug mode."""

        self.object = obj

        # TODO: remove faces_alias from mutation and pass them to builder faces?
        mutation = Mutation(
            obj,
            self.last if self.mutations else None,
            name,
            len(self.mutations),
            faces_alias,
        )

        for face in mutation.faces_added + mutation.faces_altered:
            if face.label not in self.faces_dict:
                self.faces_dict[face.label] = face

        for face in mutation.faces_added:
            self.faces_modes[face.label] = cast_mode(mode)

        for face in mutation.faces_altered:
            self.faces_modes[face.label] = AUTO

        self.mutations.append(mutation)
        return mutation

    def move(
            self,
            location: _.Location,
            mode: Mode | ColorLike = AUTO,
        ) -> Mutation:
        """Mutation: move the object to the given location, keeping the colors.
        with the given color and debug mode.
        If not color is defined, keep the previous ones for each face."""

        obj = location * self.object
        faces_alias: dict[Hash, Hash] = {}

        for face in ShapeList(self.object.faces()):
            face_moved = add_shape_hash(location * face, True)
            faces_alias[face_moved.label] = face.label

        return self.mutate('move', obj, cast_mode(mode), faces_alias)

    def add(
            self,
            part: Builder | _.Part,
            mode: Mode | ColorLike = AUTO,
        ) -> Mutation:
        """Mutation: fuse the given part to the current object.
        with the given color and debug mode."""

        obj = self.object + self._cast_part(part)
        return self.mutate('add', obj, cast_mode(mode))

    def sub(
            self,
            part: Builder | _.Part,
            mode: Mode | ColorLike = AUTO,
        ) -> Mutation:
        """Mutation: substract the given part from the current object,
        with the given color and debug mode."""

        obj = self.object - self._cast_part(part)
        return self.mutate('sub', obj, cast_mode(mode))

    def intersect(
            self,
            part: Builder | _.Part,
            mode: Mode | ColorLike = AUTO,
        ) -> Mutation:
        """Mutation: intersects the given part to the current object,
        with the given color and debug mode."""

        obj = self.object & self._cast_part(part)
        return self.mutate('inter', obj, cast_mode(mode))

    def fillet(
            self,
            edges: Iterable[_.Edge] | _.Edge,
            radius: float,
            mode: Mode | ColorLike = AUTO,
        ) -> Mutation:
        """Mutation: apply a fillet of the given radius to the given edges of
        the current object, with the given color and debug mode."""

        obj = self.object.fillet(radius, self._cast_edges(edges))
        return self.mutate('fillet', obj, cast_mode(mode))

    def chamfer(
            self,
            edges: Iterable[_.Edge] | _.Edge,
            length: float,
            length2: float | None=None,
            face: _.Face | None=None,
            mode: Mode | ColorLike = AUTO,
        ) -> Mutation:
        """Mutation: apply a chamfer of the given length to the given edges of
        the current object, with the given color and debug mode."""

        edges = self._cast_edges(edges)
        obj = self.object.chamfer(length, length2, edges, face) # type: ignore
        return self.mutate('chamfer', obj, cast_mode(mode))

    def info(self, file=None):
        """Print the list of mutations to the given file (stdout by default)."""

        palette = config.COLOR_PALETTE.build_palette(len(self.mutations))

        def row(mut: Mutation) -> tuple:
            color = palette[mut.index] # FIXME
            r, g, b = [int(c * 255) for c in color.to_tuple()[:3]]

            start = f"\033[38;2;{ r };{ g };{ b }m"
            end = "\033[0m"

            columns = {
                "idx": str(mut.index),
                "label": mut.id,
                "type": mut.name,
                "color_hex": color_to_str(color, True),
                "color_name": color_to_str(color, False),
                "f+": str(len(mut.faces_added)),
                "f~": str(len(mut.faces_altered)),
                "f-": str(len(mut.faces_removed)),
                "e+": str(len(mut.edges_added)),
                "e~": str(len(mut.edges_altered)),
                "e-": str(len(mut.edges_removed)),
            }

            return tuple(
                (f"{ start }{ col }{ end }" if config.INFO_COLOR else col)
                for header, col in columns.items()
                if header in config.COLUMNS_MUTATIONS
            )

        str_table = tabulate(
            [row(mutation) for mutation in self.mutations],
            [header.title() for header in config.COLUMNS_MUTATIONS],
            config.INFO_TABLE_FORMAT
        )
        print(str_table, file=file or stdout)

    def debug(self, faces: ShapeList, mode: Mode | ColorLike=DEBUG):
        """Set a face for debugging, so it will appear in the given color while
        the rest of the object will be translucent."""

        for face in self._cast_faces(faces):
            self.faces_modes[face.label] = cast_mode(mode)

    def export(
            self,
            exporter: _.Export2D,
            file_path: PathLike | bytes | str,
            include_part=True
        ):
        """Export the current object using the given exporter in the given file
        path. If `include_part` is false, do not include the object."""

        if include_part:
            exporter.add_shape(self.object) # type: ignore
        exporter.write(file_path) # type: ignore

    def export_stl(
            self,
            file_path: PathLike | bytes | str,
            tolerance: float = 0.001,
            angular_tolerance: float = 0.1,
            ascii_format: bool = False
        ):
        """Export the current object in STL format to the given file path,
        with the given tolerance, angular tolerance and ascii format mode."""
        _.export_stl(
            self.object,
            file_path,
            tolerance,
            angular_tolerance,
            ascii_format
        )



"""A module used to store shapes-related stuff."""
from enum import Enum
from typing import TypeAlias, Iterable, TextIO, TypeVar
from hashlib import md5
from sys import stdout

from tabulate import tabulate
import build123d as _

from . import config
from .colors import color_to_str


Hash: TypeAlias = str


class ShapeState(Enum):
    """The possible states of a shape for a mutation."""

    ADDED = 1
    ALTERED = 2
    UNTOUCHED = 3
    REMOVED = 4

def hash_shape(shape: _.Shape) -> Hash:
    """Return a reproducible hash.
    OCP 7.2 might produce better hashes that could make this unnecessary."""

    def to_int(number: float) -> int:
        return int(number * 1000)

    def serialize_vertex(vertex: _.Vertex) -> tuple:
        return tuple(to_int(v) for v in vertex.to_tuple())

    def serialize_edge(edge: _.Edge) -> tuple:
        vertices = tuple(serialize_vertex(v) for v in edge.vertices())
        is_circle = edge.geom_type == _.GeomType.CIRCLE
        radius = to_int(edge.radius) if is_circle else 0
        return (edge.geom_type, vertices, radius)

    def serialize_face(face: _.Face) -> tuple:
        return tuple(serialize_edge(edge) for edge in face.edges())

    def serialize_part(part: _.Part) -> tuple:
        return tuple(serialize_face(face) for face in part.faces())

    if isinstance(shape, _.Vertex):
        serialized = serialize_vertex(shape)
    elif isinstance(shape, _.Edge):
        serialized = serialize_edge(shape)
    elif isinstance(shape, _.Face):
        serialized = serialize_face(shape)
    elif isinstance(shape, _.Part):
        serialized = serialize_part(shape)
    else:
        raise TypeError

    return md5(str(serialized).encode()).hexdigest()


ShapeLike: TypeAlias = _.Face | _.Edge | _.Vertex
ShapeT = TypeVar("ShapeT", bound=_.Face | _.Edge | _.Vertex)


def add_shape_hash(shape: ShapeT, force=False) -> ShapeT:
    """Add the hash of the given shape to the shape label, only if it doesn't
    exit or if force is True."""
    if force or not shape.label:
        shape.label = hash_shape(shape)
    return shape


class ShapeList(_.ShapeList[ShapeT]):
    """A custom ShapeList that automatically adds a hash to the shape label,
    and with some extra utility methods."""

    def __init__(self, shapes: Iterable[ShapeT]):
        super().__init__(add_shape_hash(shape) for shape in shapes)

    def __setitem__(self, index: int, shape: ShapeT):
        super().__setitem__(index, add_shape_hash(shape))

    def insert(self, index: int, shape: ShapeT):
        super().insert(index, add_shape_hash(shape))

    def append(self, shape: ShapeT):
        super().append(add_shape_hash(shape))

    def extend(self, other: Iterable[ShapeT]):
        super().extend(
            other if isinstance(other, ShapeList)
            else [add_shape_hash(shape) for shape in other]
        )

    def get(self, shape_hash: Hash) -> ShapeT:
        """Return the shape that belongs to the given hash."""
        for shape in self:
            if shape.label == shape_hash:
                return shape
        raise KeyError(shape_hash)

    def contain(self, edge_hash: Hash) -> bool:
        """Return True if the given hash is found in the shape list."""
        for edge in self:
            if edge.label == edge_hash:
                return True
        return False

    def hashes(self) -> list[Hash]:
        """Return a list of hashes corresponding to the all shapes hash."""
        return [shape.label for shape in self]

    def info(self, file: TextIO|None=None):
        """Prints an info table of all the faces to the given file or stream
        (default to stdout)"""

        def str_vector(vector: _.Vector) -> str:
            return f"[{', '.join([f'{ n :.1g}' for n in vector.to_tuple()])}]"

        def row(shape: _.Shape) -> tuple:
            color = shape.color or config.DEFAULT_COLOR
            r, g, b = [int(c * 255) for c in color.to_tuple()[:3]]

            start = f"\033[38;2;{ r };{ g };{ b }m"
            end = "\033[0m"

            columns = {
                "hash": shape.label[:6],
                "type": shape.geom_type,
                "area": f"{ shape.area :.2g}",
                "position": str_vector(shape.position),
                "orientation": str_vector(shape.orientation),
                "color_hex": color_to_str(color, True),
                "color_name": color_to_str(color, False),
            }

            return tuple(
                (f"{ start }{ color }{ end }" if config.INFO_COLOR else color)
                for header, color in columns.items()
                if header in config.COLUMNS_SHAPES
            )

        str_table = tabulate(
            [row(shape) for shape in self],
            [header.title() for header in config.COLUMNS_SHAPES],
            config.INFO_TABLE_FORMAT
        )
        print(str_table, file=file or stdout)



"""A module used for color-related stuff."""
from enum import Enum
from typing import TypeAlias
import struct

import build123d as _
from .color_palettes import viridis, inferno, magma, plasma


ColorLike: TypeAlias = (
    _.Color | # build123d color
    _.Quantity_ColorRGBA | # OCP color
    str | # name, ex: "red"
    tuple[str, int] | # name + alpha, ex: ("red", 0.5)
    tuple[float, float, float] | # rvb, ex: (1, 0, 0)
    tuple[float, float, float, int] | # rvb + alpha, ex: (1, 0, 0, 0.5)
    int | # hexa, ex: 0xff0000
    tuple[int, int] # hexa + alpha, ex: (0xff0000, 0x80)
)


def cast_color(color: ColorLike) -> _.Color:
    """Cast a ColorLike to a Build123d Color"""
    return color if isinstance(color, _.Color) else _.Color(color)


def color_to_str(color: _.Color, as_hex=False) -> str:
    """Return a string representation of the given color."""
    def float_to_hex(n: float):
        return hex(int(n * 255))[2:].rjust(2, '0')

    if as_hex:
        r, g, b, a = [float_to_hex(c) for c in color.to_tuple()]
        return f"#{ r }{ g }{ b }{ a }"

    return str(color).split("~")[1].strip().lower()


def get_rvb(color: _.Color) -> tuple[int, int, int]:
    return color.to_tuple()[:3]


class ColorPalette(Enum):
    "The name of predefined color palettes."
    VIRIDIS = 0
    INFERNO = 1
    MAGMA = 2
    PLASMA = 3

    def build_palette(self, amount: int) -> list[_.Color]:
        """Build a list of colors based on the given palette and the amount of
        colors."""

        palette = [viridis, inferno, magma, plasma][self.value]

        def get_color(index: int) -> _.Color:
            color_int = palette[index]
            color_hex = hex(color_int)[2:].rjust(6, '0')
            color_tuple = struct.unpack('BBB', bytes.fromhex(color_hex))
            return _.Color(tuple(c/256 for c in color_tuple))

        if amount == 1:
            return [get_color(127)]

        indexes = [int(idx / (amount - 1) * 255) for idx in range(amount)]
        return [get_color(index) for index in indexes]



"""Module containing the Mutation class."""
from __future__ import annotations

import build123d as _

from .shapes import Hash, ShapeState, hash_shape, ShapeList, ShapeLike


class Mutation:
    """Class managing the mutation applied when mutating the object."""

    def __init__(
        self,
        obj: _.Part,
        previous: Mutation | None,
        name: str,
        index: int,
        faces_alias: dict[Hash, Hash] | None
    ) -> None:
        self.previous = previous
        self.name = name
        self.index = index
        self.faces_alias = faces_alias or {}

        self.id = f"{ name }-{ index }"

        self.faces = ShapeList(obj.faces())
        self.faces_state = self.get_shapes_state(_.Face)

        self.faces_added = self.filter_shapes(ShapeState.ADDED, _.Face)
        self.faces_altered = self.filter_shapes(ShapeState.ALTERED, _.Face)
        self.faces_untouched = self.filter_shapes(ShapeState.UNTOUCHED, _.Face)
        self.faces_removed = self.filter_shapes(ShapeState.REMOVED, _.Face)

        self.edges = ShapeList(obj.edges())
        self.edges_state = self.get_shapes_state(_.Edge)
        self.edges_added = self.filter_shapes(ShapeState.ADDED, _.Edge)
        self.edges_altered = self.filter_shapes(ShapeState.ALTERED, _.Edge)
        self.edges_untouched = self.filter_shapes(ShapeState.UNTOUCHED, _.Edge)
        self.edges_removed = self.filter_shapes(ShapeState.REMOVED, _.Edge)

        self.vertices = ShapeList(obj.vertices())

    def get_shapes(self, shape_type: type[ShapeLike]) -> ShapeList:
        """Return the mutation shapes belonging the given shape type."""
        if shape_type == _.Face:
            return self.faces
        if shape_type == _.Edge:
            return self.edges
        if shape_type == _.Vertex:
            return self.vertices
        raise TypeError

    def __repr__(self):
        return self.id

    def filter_shapes(self, state: ShapeState, shape_type: type[ShapeLike]) -> ShapeList:
        """Return the shapes of the current object that match the given state."""

        assert shape_type in [_.Face, _.Edge]

        mutation = (
            self.previous if self.previous and state == ShapeState.REMOVED
            else self
        )
        shapes = mutation.get_shapes(shape_type)
        shapes_state = self.faces_state if shape_type == _.Face else self.edges_state

        faces = [shapes.get(h) for h, s in shapes_state.items() if s == state]
        return ShapeList(faces)

    def get_shapes_state(self, shape_type: type[ShapeLike]) -> dict[Hash, ShapeState]:
        """Return a dictionnary holding the state of each face of the object."""

        def get_state(shape: ShapeLike) -> ShapeState:
            if not self.previous:
                return ShapeState.ADDED

            if self.previous.get_shapes(shape_type).contain(shape.label):
                return ShapeState.UNTOUCHED

            if isinstance(shape, _.Face) and self.is_altered_face(shape):
                return ShapeState.ALTERED

            if isinstance(shape, _.Edge) and self.is_altered_edge(shape):
                return ShapeState.ALTERED

            return ShapeState.ADDED

        shapes = self.get_shapes(shape_type)
        shapes_state = {shape.label: get_state(shape) for shape in shapes}

        if self.previous:
            for previous_shape in self.previous.get_shapes(shape_type):
                if not shapes.contain(previous_shape.label):
                    shapes_state[previous_shape.label] = ShapeState.REMOVED

        return shapes_state

    @classmethod
    def is_altered_faces(cls, this_face: _.Face, that_face: _.Face):
        """Check if the two given faces were altered or not, by comparing the
        edges of each face: if a similar edge is found, they are altered."""

        for this_edge in this_face.edges():
            this_hash = hash_shape(this_edge)
            for that_edge in that_face.edges():
                if this_hash == hash_shape(that_edge):
                    return True

        if (
            this_face.geom_type == that_face.geom_type
            and this_face.location == that_face.location
            and this_face.center_location == that_face.center_location
        ):
            return True

        return False

    def is_altered_face(self, face: _.Face):
        """Check if the given face were altered, by comparing the edges of the
        face: if a similar edge is found in the object, it is altered."""

        if not self.previous:
            return True

        for edge in face.edges():
            if self.previous.edges.contain(hash_shape(edge)):
                return True

        for that_face in self.previous.faces:
            if (face.geom_type == that_face.geom_type
                and face.location == that_face.location
                and face.center_location == that_face.center_location
            ):
                return True

        return False

    def is_altered_edge(self, edge: _.Edge):
        """Check if the given edge were altered, by comparing the vertices of
        the face: if a similar vertex is found in the object, it is altered."""

        if not self.previous:
            return True

        for vertex in edge.vertices():
            if self.previous.vertices.contain(hash_shape(vertex)):
                return True

        return False
